from instrumental import constructs
from instrumental.test import DummyConfig
from instrumental.test import load_module

class TestMetadataGatheringVisitor(object):
    
    def _make_one(self):
        from instrumental.metadata import MetadataGatheringVisitor as MGV
        return MGV
    
    def _get_pragmas(self, source):
        from instrumental.pragmas import PragmaFinder
        finder = PragmaFinder()
        return finder.find_pragmas(source)
    
    def test_gather_lines(self):
        def test_module():
            a = 3
            if a == 4:
                b = 0
            else:
                b = 1
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert set([1, 2, 3, 5]) == set(metadata.lines), set(metadata.lines)
    
    def test_gather_lines__with_pragmas(self):
        def test_module():
            a = 3
            if a == 4: 
                b = 0 # pragma: no cover
            else:
                b = 1
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert set([1, 2, 5]) == set(metadata.lines), set(metadata.lines)
        
    def test_gather_constructs__if_simple_decision(self):
        def test_module():
            a = 3
            if a == 4: 
                b = 0
            else:
                b = 1
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert "2.1" in metadata.constructs, metadata.constructs
        decision = metadata.constructs["2.1"]
        assert isinstance(decision, constructs.BooleanDecision)
            
    def test_gather_constructs__ifexp_simple_decision(self):
        def test_module():
            a = 3
            b = 0 if a == 4 else 1
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert "2.1" in metadata.constructs, metadata.constructs
        decision = metadata.constructs["2.1"]
        assert isinstance(decision, constructs.BooleanDecision)
            
    def test_gather_constructs__multiple_ifexps(self):
        def test_module():
            a = 3
            b, c = 0 if a == 4 else 1, 2 if a == 2 or a == 3 else 3
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert 6 == len(metadata.constructs), metadata.constructs
        decision = metadata.constructs["2.1"]
        assert isinstance(decision, constructs.BooleanDecision)
        comparison = metadata.constructs["2.2"]
        assert isinstance(comparison, constructs.Comparison)
        decision = metadata.constructs["2.3"]
        assert isinstance(decision, constructs.BooleanDecision)
        decision = metadata.constructs["2.4"]
        assert isinstance(decision, constructs.LogicalOr)
        comparison = metadata.constructs["2.5"]
        assert isinstance(comparison, constructs.Comparison)
        comparison = metadata.constructs["2.6"]
        assert isinstance(comparison, constructs.Comparison)
            
    def test_gather_constructs__while_simple_decision(self):
        def test_module():
            a = 3
            b = 0
            while a > 1:
                a -= 1
                b = 0
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert "3.1" in metadata.constructs, metadata.constructs
        decision = metadata.constructs["3.1"]
        assert isinstance(decision, constructs.BooleanDecision)

    def test_visit_Assert(self):
        def test_module():
            assert a or b
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert 2 == len(metadata.constructs)
        decision = metadata.constructs["1.1"]
        assert isinstance(decision, constructs.BooleanDecision)
        boolop = metadata.constructs["1.2"]
        assert isinstance(boolop, constructs.LogicalOr)

    def test_visit_And_unreachable_True(self):
        def test_module():
            c =  a and False
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert 2 == len(metadata.constructs), len(metadata.constructs)
        decision = metadata.constructs["1.1"]
        assert isinstance(decision, constructs.BooleanDecision)
        assert isinstance(list(decision.conditions[True])[0], constructs.UnreachableCondition)
        boolop = metadata.constructs["1.2"]
        assert isinstance(boolop, constructs.LogicalAnd)

    def test_visit_Or_unreachable_False(self):
        def test_module():
            c =  a or True
        module, source = load_module(test_module)
        pragmas = self._get_pragmas(source)
        
        metadata = self._make_one().analyze(DummyConfig(), 'modname', 'modname.py', module, pragmas)
        assert 2 == len(metadata.constructs), len(metadata.constructs)
        decision = metadata.constructs["1.1"]
        assert isinstance(list(decision.conditions[False])[0], constructs.UnreachableCondition)
        assert isinstance(decision, constructs.BooleanDecision)
        boolop = metadata.constructs["1.2"]
        assert isinstance(boolop, constructs.LogicalOr)




# This file is intended for use as a test of the instrumentation and coverage
# generation facilities. The tests should always pass.
import sys

from instrumental.samples.boolean import and_3

class TestBoolOp(object):
    
    def test_TTT(self):
        result = and_3(True, True, True)
        assert result, result
    
    def test_TFT(self):
        result = and_3(True, False, True)
        assert not result, result
    
    def test_FFT(self):
        result = and_3(False, False, True)
        assert not result, result




class TestBooleanEvaluator(object):
    
    def _test_node(self, node, expected):
        from instrumental.metadata import BooleanEvaluator
        result = BooleanEvaluator.evaluate(node)
        assert expected == result, (expected, result)
    
    def test_BoolOp_2_pin_and_without_literals(self):
        from astkit import ast
        node = ast.BoolOp(op=ast.And(),
                          values=[ast.Name(id='a'),
                                  ast.Name(id='b'),
                                  ],
                          )
        expected = set([True, False])
        yield self._test_node, node, expected
    
    def test_BoolOp_3_pin_and_without_literals(self):
        from astkit import ast
        node = ast.BoolOp(op=ast.And(),
                          values=[ast.Name(id='a'),
                                  ast.Name(id='b'),
                                  ast.Name(id='c'),
                                  ],
                          )
        expected = set([True, False])
        yield self._test_node, node, expected
    
    def test_BoolOp_3_pin_and_with_1_literal_True(self):
        from astkit import ast
        node = ast.BoolOp(op=ast.And(),
                          values=[ast.Name(id='a'),
                                  ast.Name(id='True'),
                                  ast.Name(id='c'),
                                  ],
                          )
        expected = set([True, False])
        yield self._test_node, node, expected
    
    def test_BoolOp_3_pin_and_with_1_literal_False(self):
        from astkit import ast
        node = ast.BoolOp(op=ast.And(),
                          values=[ast.Name(id='a'),
                                  ast.Name(id='False'),
                                  ast.Name(id='c'),
                                  ],
                          )
        expected = set([False])
        yield self._test_node, node, expected
    
    def test_BoolOp_2_pin_or_without_literals(self):
        from astkit import ast
        node = ast.BoolOp(op=ast.Or(),
                          values=[ast.Name(id='a'),
                                  ast.Name(id='b'),
                                  ],
                          )
        expected = set([True, False])
        yield self._test_node, node, expected
    
    def test_BoolOp_3_pin_or_without_literals(self):
        from astkit import ast
        node = ast.BoolOp(op=ast.Or(),
                          values=[ast.Name(id='a'),
                                  ast.Name(id='b'),
                                  ast.Name(id='c'),
                                  ],
                          )
        expected = set([True, False])
        yield self._test_node, node, expected
    
    def test_BoolOp_3_pin_or_with_1_literal_True(self):
        from astkit import ast
        node = ast.BoolOp(op=ast.Or(),
                          values=[ast.Name(id='a'),
                                  ast.Name(id='True'),
                                  ast.Name(id='c'),
                                  ],
                          )
        expected = set([True])
        yield self._test_node, node, expected
    
    def test_BoolOp_3_pin_or_with_1_literal_False(self):
        from astkit import ast
        node = ast.BoolOp(op=ast.Or(),
                          values=[ast.Name(id='a'),
                                  ast.Name(id='False'),
                                  ast.Name(id='c'),
                                  ],
                          )
        expected = set([True, False])
        yield self._test_node, node, expected
    
    def test_IfExp_without_literals(self):
        from astkit import ast
        test = ast.Name(id='a')
        body = ast.Name(id='b')
        orelse = ast.Name(id='c')
        node = ast.IfExp(test=test, body=body, orelse=orelse)
        expected = set([True, False])
        yield self._test_node, node, expected

    def test_IfExp_with_literal_non_literal_test(self):
        from astkit import ast
        test = ast.Name(id='a')
        body = ast.Num(n=4)
        orelse = ast.Name(id='c')
        node = ast.IfExp(test=test, body=body, orelse=orelse)
        expected = set([True, False])
        yield self._test_node, node, expected

    def test_IfExp_with_literal_body_truthy(self):
        from astkit import ast
        test = ast.Name(id='True')
        body = ast.Num(n=4)
        orelse = ast.Name(id='c')
        node = ast.IfExp(test=test, body=body, orelse=orelse)
        expected = set([True])
        yield self._test_node, node, expected

    def test_IfExp_with_literal_body_falsy(self):
        from astkit import ast
        test = ast.Name(id='True')
        body = ast.Num(n=0)
        orelse = ast.Name(id='c')
        node = ast.IfExp(test=test, body=body, orelse=orelse)
        expected = set([False])
        yield self._test_node, node, expected

    def test_IfExp_with_literal_orelse_truthy(self):
        from astkit import ast
        test = ast.Name(id='False')
        body = ast.Name(id='b')
        orelse = ast.Num(n=4)
        node = ast.IfExp(test=test, body=body, orelse=orelse)
        expected = set([True])
        yield self._test_node, node, expected

    def test_IfExp_with_literal_orelse_falsy(self):
        from astkit import ast
        test = ast.Name(id='False')
        body = ast.Name(id='b')
        orelse = ast.Num(n=0)
        node = ast.IfExp(test=test, body=body, orelse=orelse)
        expected = set([False])
        yield self._test_node, node, expected

    def test_UnaryOp_Not_variable(self):
        from astkit import ast
        node = ast.UnaryOp(op=ast.Not(), operand=ast.Name(id='a'))
        expected = set([True, False])
        yield self._test_node, node, expected
        
    def test_UnaryOp_Not_literal_True(self):
        from astkit import ast
        node = ast.UnaryOp(op=ast.Not(), operand=ast.Name(id='True'))
        expected = set([False])
        yield self._test_node, node, expected
        
    def test_UnaryOp_Not_literal_False(self):
        from astkit import ast
        node = ast.UnaryOp(op=ast.Not(), operand=ast.Name(id='False'))
        expected = set([True])
        yield self._test_node, node, expected
        
    def test_UnaryOp_UAdd(self):
        from astkit import ast
        node = ast.UnaryOp(op=ast.UAdd(), operand=ast.Num(n=4))
        expected = set([True, False])
        yield self._test_node, node, expected
    
    def test_List_truthy(self):
        from astkit import ast
        node = ast.List(elts=[ast.Name(id='False'),
                              ])
        expected = set([True])
        yield self._test_node, node, expected
        
    def test_List_falsy(self):
        from astkit import ast
        node = ast.List(elts=[])
        expected = set([False])
        yield self._test_node, node, expected
        
    def test_True(self):
        from astkit import ast
        node = ast.Lambda(args=ast.arguments(), body=ast.Name(id='a'))
        expected = set([True])
        yield self._test_node, node, expected



import sys

def test_func(a, b, c, d, e, i=5):
    
    if a and b:
        sys.stdout.write('11')
    
    if b or c:
        sys.stdout.write('12')
    
    while i > 0:
        i -= 1
    
    x = 4 if d else 7
    
    y = e and True or 16
    



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from copy import deepcopy

from astkit import ast
from astkit.render import SourceCodeRenderer

class PragmaCondition(object):
    TAG = 'P'
    
    def __str__(self):
        return self.TAG

class UnreachableCondition(object):
    TAG = 'U'
    
    def __str__(self):
        return self.TAG

class LogicalBoolean(object):
    
    def __init__(self, modulename, label, node, pragmas):
        self.modulename = modulename
        self.label = label
        self.node = deepcopy(node)
        self.pragmas = pragmas
        self.source = SourceCodeRenderer.render(node)
        self.pins = len(node.values)
        self.conditions =\
            dict((i, set()) for i in range(self.pins + 1))
        self.literals = self._gather_literals(node)
        self._set_unreachable_conditions()
        for pragma in pragmas:
            if hasattr(pragma, 'selector'):
                pragma_label = '%s.%s' % (node.lineno, pragma.selector)
                if label == pragma_label:
                    pragma.apply(self)
    
    def _gather_literals(self, node):
        _literals = {}
        for i, value in enumerate(node.values):
            # Try to determine if the condition is a literal
            # Maybe we can do something with this information?
            try:
                literal = ast.literal_eval(value)
                _literals[i] = literal
            except ValueError:
                pass
        return _literals
    
    def _set_unreachable_conditions(self):
        for condition in self.unreachable_conditions():
            self.conditions[condition] = UnreachableCondition()
    
    @property
    def lineno(self):
        return self.node.lineno
    
    def is_decision(self):
        return False
    
    def number_of_conditions(self, report_conditions_with_literals):
        if report_conditions_with_literals:
            return len(self.conditions)
        else:
            return len(self.conditions) - len(self.unreachable_conditions())
    
    def number_of_conditions_hit(self):
        return len([value 
                    for value in self.conditions.values()
                    if value and not isinstance(value, UnreachableCondition)])
    
    def conditions_missed(self, report_conditions_with_literals):
        if report_conditions_with_literals:
            missed = [self.description(n) for n in self.conditions
                      if (not self.conditions[n] 
                          or isinstance(self.conditions[n], UnreachableCondition))]
        else:
            missed = [self.description(n) for n in self.conditions
                      if not (isinstance(self.conditions[n], UnreachableCondition)
                              or self.conditions[n])]
        return missed
    
    def _literal_warning(self):
        return ("** One or more condition combinations may not be reachable due"
                " to the presence of a literal in the decision")
    
    def _format_condition_result(self, result, length=6):
        if isinstance(result, UnreachableCondition):
            return str(result)
        else:
            padding = '\n' + (' ' * length)
            return padding.join(tag for tag in sorted(result))
    
    def result(self):
        lines = []
        name = "%s -> %s:%s < %s >" % (self.__class__.__name__, self.modulename, self.label, self.source)
        lines.append("%s" % (name,))
        if self.literals:
            lines.append("")
            lines.append(self._literal_warning())
        lines.append("")
        for condition in sorted(self.conditions):
            desc = self.description(condition)
            length = len(desc) + 5
            lines.append(desc +
                         " ==> " + self._format_condition_result(self.conditions[condition],
                                                                 length=length))
        return "\n".join(lines)
    
    def decision_result(self):
        lines = []
        name = "Decision -> %s:%s < %s >" % (self.modulename, self.label, self.source,)
        lines.append("%s" % (name,))
        lines.append("")
        lines.append("T ==> %s" % self._format_condition_result(self.was_true()))
        lines.append("F ==> %s" % self._format_condition_result(self.was_false()))
        return "\n".join(lines)
    

class LogicalAnd(LogicalBoolean):
    """ Stores the execution information for a Logical And
        
        For an and condition with n inputs, there will be
        n + 2 recordable conditions. Condition 0 indicates
        that all inputs are True. Conditions 1 though n
        indicate that the input in the numbered position
        is False and all inputs before it are True. All
        inputs after the numbered position are, in this
        case, considered to be "don't care" since they will
        never be evaluated.
    """
    
    def unreachable_conditions(self):
        """ Return the unreachable conditions for this construct
            
            Note that conditions are reported as unreachable if they cannot
            happen due to the presence of literals in the expression *and*
            if they do not represent a situation in which later inputs are
            left unevaluated because of a literal value earlier in the
            expression.
        """
        _conditions = []
        if not self.literals.get(self.pins-1, True):
            _conditions.append(0)
        for i, literal in self.literals.items():
            if literal:
                _conditions.append(i+1)
        return _conditions
    
    def record(self, value, pin, tag):
        """ Record that a value was seen for a particular pin """
        # If the pin is not the last pin in the decision and
        # the value seen is False, then we've found the pin
        # that has forced the decision False and we should
        # record that.
        if pin < (self.pins-1):
            if not value:
                self.conditions[pin+1].add(tag)
        
        # If the pin is the last pin then we'll record that
        # it either allowed the decision to be True or it
        # is the pin that has forced the decision False.
        elif pin == (self.pins-1):
            if value:
                self.conditions[0].add(tag)
            else:
                self.conditions[self.pins].add(tag)
    
    def description(self, n):
        if n == 0:
            return " ".join("T" * self.pins)
        elif n < (self.pins + 1):
            acc = ""
            if n > 1:
                acc += " ".join("T" * (n - 1)) + " "
            acc += "F"
            if self.pins - n:
                acc += " " + " ".join("*" * (self.pins - n))
            return acc
        elif n == (self.pins + 1):
            return "Other"
    
    def was_true(self):
        return self.conditions[0]
    
    def was_false(self):
        results = set()
        for n in range(1, self.pins+1):
            results.update(self.conditions[n])
        return results
    
class LogicalOr(LogicalBoolean):
    """ Stores the execution information for a Logical Or
        
        For an or condition with n inputs, there will be
        n + 2 recordable conditions. Conditions 0 though
        n indicate that the numbered position input is
        True. Condition n indicates that all inputs are
        False. Condition n + 1 is "Other".
    """
    
    def unreachable_conditions(self):
        """ Return the unreachable conditions for this construct
            
            Note that conditions are reported as unreachable if they cannot
            happen due to the presence of literals in the expression *and*
            if they do not represent a situation in which later inputs are
            left unevaluated because of a literal value earlier in the
            expression.
        """
        _conditions = []
        for i, literal in self.literals.items():
            if not literal:
                _conditions.append(i)
        if self.literals.get(self.pins-1):
            _conditions.append(self.pins)
        return _conditions
    
    def record(self, value, pin, tag):
        """ Record that a value was seen for a particular pin """
        
        # If the pin is not the last pin in the decision
        # and the value we see is True, then we've found the
        # condition that forces the decision True in this case.
        # If the value we see is False then we'll ignore it
        # since this is not a significant case.
        if pin < (self.pins-1):
            if value:
                self.conditions[pin].add(tag)
        
        # If this is the last pin then it either allowed the
        # decision to be False or forced the decision True.
        elif pin == (self.pins-1):
            if value:
                self.conditions[pin].add(tag)
            else:
                self.conditions[self.pins].add(tag)
        
    def description(self, n):
        acc = ""
        if n < self.pins:
            if n > 0:
                acc += " ".join("F" * n) + " "
            acc += "T"
            if self.pins - n - 1:
                acc += " " + " ".join("*" * (self.pins - n - 1))
            return acc
        elif n == self.pins:
            return " ".join("F" * self.pins)
        elif n == (self.pins + 1):
            return "Other"
    
    def was_true(self):
        results = set()
        for n in range(0, self.pins):
            results.update(self.conditions[n])
        return results
    
    def was_false(self):
        return self.conditions[self.pins]
    
class BooleanDecision(object):
    
    def __init__(self, modulename, label, node, pragmas):
        self.modulename = modulename
        self.label = label
        self.node = deepcopy(node)
        self.pragmas = pragmas
        self.lineno = node.lineno
        self.source = SourceCodeRenderer.render(node)
        self.conditions = {True: set(),
                           False: set()}
        for pragma in pragmas:
            if hasattr(pragma, 'selector'):
                pragma_label = '%s.%s' % (node.lineno, pragma.selector)
                if label == pragma_label:
                    pragma.apply(self)
    
    def is_decision(self):
        return True
    
    def record(self, expression, tag):
        result = bool(expression)
        self.conditions[result].add(tag)
        return result
    
    def description(self, condition):
        return str(bool(condition))[0]
    
    def was_true(self):
        return self.conditions[True]
    
    def was_false(self):
        return self.conditions[False] 
    
    def set_unreachable(self, condition):
        self.conditions[condition].add(UnreachableCondition())
    
    def number_of_conditions(self, report_conditions_with_literals):
        if report_conditions_with_literals:
            return len(self.conditions)
        
        unreachable_conditions = 0
        for condition in self.conditions:
            for result in self.conditions[condition]:
                if (isinstance(result, UnreachableCondition)
                    or result == PragmaCondition.TAG):
                    unreachable_conditions += 1
                    break
        return len(self.conditions) - unreachable_conditions
    
    def number_of_conditions_hit(self):
        def is_hit(results):
            if results:
                return any(not (isinstance(result, UnreachableCondition)
                                or result == PragmaCondition.TAG)
                           for result in results)
            return False
        
        return len([value 
                    for value in self.conditions.values() 
                    if is_hit(value)])
    
    def conditions_missed(self, report_conditions_with_literals):
        return self.number_of_conditions(report_conditions_with_literals) - self.number_of_conditions_hit()
    
    def _format_condition_result(self, result, length=6):
        padding = '\n' + (' ' * length)
        return padding.join(str(tag) for tag in sorted(result))
    
    def result(self):
        lines = []
        name = "Decision -> %s:%s < %s >" % (self.modulename, self.label, self.source)
        lines.append("%s" % (name,))
        lines.append("")
        lines.append("T ==> %s" % self._format_condition_result(self.conditions[True]))
        lines.append("F ==> %s" % self._format_condition_result(self.conditions[False]))
        return "\n".join(lines)

class Comparison(object):
    
    def __init__(self, modulename, label, node, pragmas):
        self.modulename = modulename
        self.label = label
        self.node = deepcopy(node)
        self.pragmas = pragmas
        self.lineno = node.lineno
        self.source = SourceCodeRenderer.render(node)
        self.conditions = {True: set(),
                           False: set()}
        self._set_unreachable_condition()
        
        for pragma in pragmas:
            if hasattr(pragma, 'selector'):
                pragma_label = '%s.%s' % (node.lineno, pragma.selector)
                if label == pragma_label:
                    pragma.apply(self)
    
    def _set_unreachable_condition(self):
        self._unreachable_condition = self.unreachable_condition()
        if self._unreachable_condition is not None:
            self.conditions[self._unreachable_condition].add(
                UnreachableCondition())
    
    def unreachable_condition(self):
        try:
            _left = ast.literal_eval(self.node.left)
            _right = ast.literal_eval(self.node.comparators[0])
            if isinstance(self.node.ops[0], ast.Eq):
                condition = _left != _right
            elif isinstance(self.node.ops[0], ast.NotEq):
                condition = _left == _right
            elif isinstance(self.node.ops[0], ast.Lt):
                condition = _left >= _right
            elif isinstance(self.node.ops[0], ast.LtE):
                condition = _left > _right
            elif isinstance(self.node.ops[0], ast.Gt):
                condition = _left <= _right
            elif isinstance(self.node.ops[0], ast.GtE):
                condition = _left < _right
            elif isinstance(self.node.ops[0], ast.Is):
                condition = _left is not _right
            elif isinstance(self.node.ops[0], ast.IsNot):
                condition = _left is _right
            elif isinstance(self.node.ops[0], ast.In):
                condition = _left not in _right
            elif isinstance(self.node.ops[0], ast.NotIn):
                condition = _left in _right
            return condition
        except ValueError as exc:
            print exc
    
    def is_decision(self):
        return False
    
    def record(self, expression, tag):
        result = bool(expression)
        self.conditions[result].add(tag)
        return result
    
    def description(self, condition):
        return str(bool(condition))[0]
    
    def was_true(self):
        return self.conditions[True]
    
    def was_false(self):
        return self.conditions[False] 
    
    def set_unreachable(self, condition):
        self.conditions[condition].add(UnreachableCondition())
    
    def number_of_conditions(self, report_conditions_with_literals):
        if report_conditions_with_literals:
            return len(self.conditions)
        
        print self.conditions
        unreachable_conditions = 0
        for condition in self.conditions:
            for result in self.conditions[condition]:
                if (isinstance(result, UnreachableCondition)
                    or result == PragmaCondition.TAG):
                    unreachable_conditions += 1
                    break
        return len(self.conditions) - unreachable_conditions
    
    def number_of_conditions_hit(self):
        def is_hit(results):
            if results:
                return any(not (isinstance(result, UnreachableCondition)
                                or result == PragmaCondition.TAG)
                           for result in results)
            return False
        
        return len([value 
                    for value in self.conditions.values() 
                    if is_hit(value)])
    
    def conditions_missed(self, report_conditions_With_literals):
        return self.number_of_conditions(report_conditions_With_literals) - self.number_of_conditions_hit()
    
    def _format_condition_result(self, result, length=6):
        padding = '\n' + (' ' * length)
        return padding.join(str(tag) for tag in sorted(result))
    
    def result(self):
        lines = []
        name = "Compare -> %s:%s < %s >" % (self.modulename, self.label, self.source)
        lines.append("%s" % (name,))
        lines.append("")
        lines.append("T ==> %s" % self._format_condition_result(self.conditions[True]))
        lines.append("F ==> %s" % self._format_condition_result(self.conditions[False]))
        return "\n".join(lines)



import sys

from instrumental import importer

class FakeRecorder(object):
    pragmas = {}
    
    def add_source(self, name, source):
        pass

class FakeSys(object):
    modules = {}

class FakeVisitor(object):
    
    def visit(self, node):
        return node

class FakeVisitorFactory(object):
    recorder = FakeRecorder()
    
    def create(self, name, source):
        visitor = FakeVisitor()
        return visitor

class TestModuleLoader(object):
    
    def setup(self):
        self.path = __file__
        if self.path.endswith('.pyc'):
            self.path = self.path[:-1]
        self.name = __name__
        self._fake_sys = FakeSys()
        self._old_sys = importer.sys
        importer.sys = self._fake_sys
    
    def teardown(self):
        importer.sys = self._old_sys
    
    def _makeOne(self):
        visitor_factory = FakeVisitorFactory()
        return importer.ModuleLoader(self.path, visitor_factory)
    
    def test_load_module(self):
        loader = self._makeOne()
        module = loader.load_module(self.name)
        assert module
        assert not hasattr(module, 'path')
        assert module.__name__ in self._fake_sys.modules

class TestImportHook(object):
    
    def _makeOne(self):
        self.name = __name__
        visitor_factory = FakeVisitorFactory()
        return importer.ImportHook(self.name, visitor_factory)
    
    # def test_find_module(self):
    #     hook = self._makeOne()
    #     loader = hook.find_module(self.name)
    #     assert loader



from setuptools import setup, find_packages

version = '0.4b1'

setup(name='instrumental',
      version=version,
      description="A condition/decision coverage tool for Python",
      long_description="""\
Instrumental automatically instruments your code to add function calls that
reveal execution characteristics. Instrumental can then monitor
execution of your code and report back to you on how your source was executed.

When you run a script with Instrumental, it will tell you which decisions and
conditions haven't been fully exercised. So for a decision in an if statement,
it will tell you if the decision only ever evaluated True or False. Instrumental
will also tell you if the conditions in boolean decisions (and, or) were ever
executed both True and False.

As an example: if you usually say

python setup.py nosetests

you can say

instrumental -rS -t <packagename> setup.py nosetests

where packagename is the name of your project's top-level package.

For more information visit `the Instrumental documentation <http://instrumental.readthedocs.org//en/latest>`_.

Announcements and discussion about instrumental can be found on the
"python-instrumental-discuss" google group.

Please file bug reports and enhancement requests `on bitbucket <http://bitbucket.org/desmaj/instrumental>`_.

""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='ast condition decision coverage',
      author='Matthew J Desmarais',
      author_email='matthew.desmarais@gmail.com',
      url='',
      license='GPL',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'astkit>=0.4',
        'mako',
        ],
      tests_require=[
        'fudge',
        ],
      entry_points="""
      [console_scripts]
      instrumental = instrumental.run:main
      [nose.plugins]
      instrumental-tag = instrumental.plugins.nose.tag_test_cases:InstrumentalTagPlugin
      """,
      )



import logging

# REMIND: Can't import instrumental here or it'll throw off coverage results
#         when using the nose coverage plugin since coverage isn't setup yet.
# from instrumental.recorder import ExecutionRecorder

from nose.plugins import Plugin

class InstrumentalTagPlugin(Plugin):
    name = 'instrumental-tag'
    
    def help(self):
        return 'Tags each test case for use with instrumental'
    
    def startTest(self, test):
        from instrumental.api import Coverage
        if hasattr(test, 'address'):
            tag = ':'.join(test.address()[1:])
        else:
            tag = test.id()
        cov = Coverage()
        cov.start_context(tag)
    
    def stopTest(self, test):
        from instrumental.api import Coverage
        cov = Coverage()
        cov.stop_context()



from instrumental.importer import ImportHook

class TestImportHook(object):
    
    def setup(self):
        self.visitor_factory = object()
        self.recorder = object()
    
    def _makeOne(self, target, ignores=[]):
        return ImportHook(target, ignores, self.visitor_factory)
    
    def test_exactly_matching_target(self):
        hook = self._makeOne('pyramid')
        
    



from instrumental import summary
from instrumental.test import DummyConfig

class TestPackage(object):
    
    def test_tests(self):
        assert 'instrumental.tests' == summary._package('instrumental.tests')
    
    def test_summary(self):
        assert 'instrumental' == summary._package('instrumental.summary')

class DummyConstruct(object):
    
    def __init__(self, modulename, decision, conditions):
        self.modulename = modulename
        self._decision = decision
        self.conditions = conditions
    
    def is_decision(self):
        return self._decision

class TestExecutionSummary(object):
    
    def test_empty(self):
        assert summary.ExecutionSummary({}, {}, None)
    
    def test_nonempty(self):
        statements = {'instrumental': {1: True, 2: True, 4: False},
                      'instrumental.tests': {1: False, 3:True, 5:False},
                      }
        _1_1 = DummyConstruct('instrumental',
                              True, 
                              {True: set(['X']),
                               False: set(['X']),
                               })
        _1_2 = DummyConstruct('instrumental',
                              False,
                              {0: set(['X']),
                               1: set(),
                               2: set(),
                               })
        _3_1 = DummyConstruct('intrumental.test',
                              True,
                              {True: set(['X']),
                               False: set(['X']),
                               })
        _3_2 = DummyConstruct('instrumental.test',
                              False,
                              {0: set(['X']),
                               1: set(),
                               2: set(),
                               })
        conditions = {'instrumental': 
                      {'1.1': _1_1,
                       '1.2': _1_2,
                       },
                      'instrumental.tests': 
                      {'3.1': _3_1,
                       '3.2': _3_2,
                       }
                      }
        s = summary.ExecutionSummary(conditions, statements, DummyConfig())
        packages = s.packages
        assert 'instrumental' in packages



import ast

from instrumental.constructs import LogicalAnd
from instrumental.constructs import LogicalBoolean
from instrumental.constructs import LogicalOr
from instrumental.recorder import ExecutionRecorder
from instrumental.test import DummyConfig

class ThreePinTestCase(object):
    
    def setup(self):
        self.modulename = 'somename'
        self.node = ast.BoolOp(values=[ast.Name(id='a'),
                                       ast.Name(id='b'),
                                       ast.Name(id='c'),
                                       ],
                               lineno=6,
                               col_offset=4)
        self.label = '6.1'

class TwoPinTestCase(object):
    
    def setup(self):
        self.modulename = 'somename'
        self.node = ast.BoolOp(values=[ast.Name(id='a'),
                                       ast.Name(id='b'),
                                       ],
                               lineno=6,
                               col_offset=4)
        self.label = '6.1'

class TestLogicalAnd(ThreePinTestCase):
    
    def setup(self):
        super(TestLogicalAnd, self).setup()
        self.node.op = ast.And()
    
    def _makeOne(self):
        return LogicalAnd(self.modulename, self.label, self.node, set())
    
    def test_constructor(self):
        assert LogicalAnd(self.modulename, self.label, self.node, set())
    
    def test_has_conditions(self):
        assert hasattr(self._makeOne(), 'conditions')
    
    def test_has_n_plus_2_conditions(self):
        and_ = self._makeOne()
        assert 4 == len(and_.conditions)
    
    def test_has_conditions_0_through_n_plus_1(self):
        and_ = self._makeOne()
        for i in range(4):
            assert i in and_.conditions
    
    def test_condition_0_is_all_true(self):
        and_ = self._makeOne()
        assert "T T T" == and_.description(0)
        
    def test_condition_1_is_first_pin_false(self):
        and_ = self._makeOne()
        assert "F * *" == and_.description(1), and_.description(1)
        
    def test_condition_2_is_second_pin_false(self):
        and_ = self._makeOne()
        assert "T F *" == and_.description(2), and_.description(2)
        
    def test_condition_3_is_third_pin_false(self):
        and_ = self._makeOne()
        assert "T T F" == and_.description(3), and_.description(3)
    
    def _expect_result(self, *set_conditions):
        expected_result = "\n".join(["LogicalAnd -> somename:6.1 < (a and b and c) >",
                                     "",
                                     "T T T ==> %s" % ('X' if "T T T" in set_conditions else ''),
                                     "F * * ==> %s" % ('X' if "F * *" in set_conditions else ''),
                                     "T F * ==> %s" % ('X' if "T F *" in set_conditions else ''),
                                     "T T F ==> %s" % ('X' if "T T F" in set_conditions else ''),
                                     ])
        return expected_result
                                     
    def test_all_true(self):
        and_ = self._makeOne()
        and_.record(True, 0, 'X')
        and_.record(True, 1, 'X')
        and_.record(True, 2, 'X')
        expected_result = self._expect_result('T T T')
        assert expected_result == and_.result()
        
    def test_F_T_T(self):
        and_ = self._makeOne()
        and_.record(False, 0, 'X')
        expected_result = self._expect_result('F * *')
        assert expected_result == and_.result(), and_.result()
        
    def test_T_F_T(self):
        and_ = self._makeOne()
        and_.record(True, 0, 'X')
        and_.record(False, 1, 'X')
        expected_result = self._expect_result('T F *')
        assert expected_result == and_.result()
        
    def test_T_T_F(self):
        and_ = self._makeOne()
        and_.record(True, 0, 'X')
        and_.record(True, 1, 'X')
        and_.record(False, 2, 'X')
        expected_result = self._expect_result('T T F')
        assert expected_result == and_.result()
        
class TestLogicalAnd2Pin(TwoPinTestCase):
    
    def setup(self):
        super(TestLogicalAnd2Pin, self).setup()
        self.node.op = ast.And()
    
    def _makeOne(self):
        return LogicalAnd(self.modulename, self.label, self.node, set())
    
    def test_2_pin_and_condition_0(self):
        and_ = self._makeOne()
        assert "T T" == and_.description(0)

    def test_2_pin_and_condition_1(self):
        and_ = self._makeOne()
        assert "F *" == and_.description(1)

    def test_2_pin_and_condition_2(self):
        and_ = self._makeOne()
        assert "T F" == and_.description(2)

    def test_2_pin_and_condition_3(self):
        and_ = self._makeOne()
        assert "Other" == and_.description(3)

class TestLogicalOr(ThreePinTestCase):
    
    def setup(self):
        super(TestLogicalOr, self).setup()
        self.node.op = ast.Or()
    
    def _makeOne(self):
        return LogicalOr(self.modulename, self.label, self.node, set())
    
    def test_has_conditions(self):
        or_ = self._makeOne()
        assert hasattr(or_, 'conditions')
    
    def test_has_n_plus_1_conditions(self):
        or_ = self._makeOne()
        assert 4 == len(or_.conditions)
    
    def test_has_conditions_0_through_n_plus_1(self):
        or_ = self._makeOne()
        for i in range(4):
            assert i in or_.conditions, (i, or_.conditions)
    
    def test_condition_0_is_first_pin_true(self):
        or_ = self._makeOne()
        assert "T * *" == or_.description(0), or_.description(0)
        
    def test_condition_1_is_second_pin_true(self):
        or_ = self._makeOne()
        assert "F T *" == or_.description(1), or_.description(1)
        
    def test_condition_2_is_third_pin_true(self):
        or_ = self._makeOne()
        assert "F F T" == or_.description(2), or_.description(2)
        
    def test_condition_3_is_all_false(self):
        or_ = self._makeOne()
        assert "F F F" == or_.description(3), or_.description(3)
        
    def test_condition_4_is_other(self):
        or_ = self._makeOne()
        assert "Other" == or_.description(4)
    
    def _expect_result(self, *set_conditions):
        expected_result = "\n".join(["LogicalOr -> somename:6.1 < (a or b or c) >",
                                     "",
                                     "T * * ==> %s" % ('X' if "T * *" in set_conditions else ''),
                                     "F T * ==> %s" % ('X' if "F T *" in set_conditions else ''),
                                     "F F T ==> %s" % ('X' if "F F T" in set_conditions else ''),
                                     "F F F ==> %s" % ('X' if "F F F" in set_conditions else ''),
                                     ])
        return expected_result
                                     
    def test_T_F_F(self):
        or_ = self._makeOne()
        or_.record(True, 0, 'X')
        expected_result = self._expect_result('T * *')
        assert expected_result == or_.result(), or_.result()
        
    def test_F_T_F(self):
        or_ = self._makeOne()
        or_.record(False, 0, 'X')
        or_.record(True, 1, 'X')
        expected_result = self._expect_result('F T *')
        assert expected_result == or_.result()
        
    def test_F_F_T(self):
        or_ = self._makeOne()
        or_.record(False, 0, 'X')
        or_.record(False, 1, 'X')
        or_.record(True, 2, 'X')
        expected_result = self._expect_result('F F T')
        assert expected_result == or_.result()
        
    def test_F_F_F(self):
        or_ = self._makeOne()
        or_.record(False, 0, 'X')
        or_.record(False, 1, 'X')
        or_.record(False, 2, 'X')
        expected_result = self._expect_result('F F F')
        assert expected_result == or_.result()
        
class TestLogicalOr2Pin(TwoPinTestCase):
    
    def setup(self):
        super(TestLogicalOr2Pin, self).setup()
        self.node.op = ast.Or()
    
    def _makeOne(self):
        return LogicalOr(self.modulename, self.label, self.node, set())
    
    def test_2_pin_or_condition_0(self):
        or_ = self._makeOne()
        assert "T *" == or_.description(0)

    def test_2_pin_or_condition_1(self):
        or_ = self._makeOne()
        assert "F T" == or_.description(1)

    def test_2_pin_or_condition_2(self):
        or_ = self._makeOne()
        assert "F F" == or_.description(2)

    def test_2_pin_or_condition_3(self):
        or_ = self._makeOne()
        assert "Other" == or_.description(3)

class TestLogicalBoolean(object):
    
    def setup(self):
        self.modulename = 'somename.subname'
        self.node = ast.BoolOp(values=[ast.Name(id='a'),
                                       ast.Name(id='b')],
                               op=ast.And(),
                               lineno=17,
                               col_offset=24,
                               )
        self.label = '17.1'
    
    def _makeOne(self):
        return LogicalAnd(self.modulename, self.label, self.node, set())
    
    def test_modulename(self):
        construct = self._makeOne()
        assert self.modulename == construct.modulename
    
    def test_pins(self):
        construct = self._makeOne()
        assert 2 == construct.pins
    
    def test_source(self):
        construct = self._makeOne()
        assert "(a and b)" == construct.source
    
    def test_lineno(self):
        construct = self._makeOne()
        assert 17 == construct.lineno
    
    def test_conditions(self):
        construct = self._makeOne()
        assert {0: set(),
                1: set(),
                2: set(),
                } == construct.conditions, construct.conditions
    
    def test_number_of_conditions(self):
        construct = self._makeOne()
        assert 3 == construct.number_of_conditions(False)
    
    def test_number_of_conditions_hit(self):
        construct = self._makeOne()
        assert 0 == construct.number_of_conditions_hit()
    
    def test_conditions_missed(self):
        construct = self._makeOne()
        assert 3 == len(construct.conditions_missed(False))
    
    def test_initial_result(self):
        construct = self._makeOne()
        expected_result = "\n".join(["LogicalAnd -> somename.subname:17.1 < (a and b) >",
                                     "",
                                     "T T ==> ",
                                     "F * ==> ",
                                     "T F ==> ",
                                     ])
        assert expected_result == construct.result()

class TestLiteralInConstruct(object):
    
    def setup(self):
        # Reset recorder
        self.modulename = 'somename'
        self.source = """




x = a or None"""
        self.label = '6.2'
    
    def _makeOne(self):
        from instrumental.metadata import MetadataGatheringVisitor
        metadata = MetadataGatheringVisitor.analyze(DummyConfig(),
                                                    self.modulename,
                                                    'somefile.py',
                                                    self.source, 
                                                    {6: set()})
        return metadata.constructs[self.label]
        
    def test_presence_of_a_literal(self):
        construct = self._makeOne()
        assert "literal" in construct.result(), construct.result()
    
    def test_unreachable_condition(self):
        from instrumental.constructs import UnreachableCondition
        construct = self._makeOne()
        assert 2 == construct.number_of_conditions(False)
        assert not construct.conditions[0]
        assert isinstance(construct.conditions[1], UnreachableCondition)
        assert not construct.conditions[2]
        assert set(['T *', 'F F']) == set(construct.conditions_missed(False))
    
    def test_count_literals(self):
        from instrumental.constructs import UnreachableCondition
        construct = self._makeOne()
        assert 3 == construct.number_of_conditions(True)
        assert not construct.conditions[0]
        assert isinstance(construct.conditions[1], UnreachableCondition)
        assert not construct.conditions[2]
        assert set(['T *', 'F T', 'F F']) == set(construct.conditions_missed(True))



from instrumental.reporting import ExecutionReport

class FakeMetadata(object):
    
    def __init__(self, constructs, statements=[]):
        self.constructs = constructs
        self.statements = statements
    
class FakeConstruct(object):
    
    def __init__(self, label, *conditions_hit):
        self.label = label
        self.modulename = "somemodulename"
        self.lineno = 17
        self.conditions = {True: set(),
                           False: set()}
        for condition in conditions_hit:
            self.conditions[condition].add('*')
    
    def is_decision(self):
        return False
    
    def result(self):
        return "ConstructResult(%s)" % self.label
    
class TestExecutionReport(object):
    
    def setup(self):
        self.T = FakeConstruct("True only", True)
        self.F = FakeConstruct("False only", False)
        self.TF = FakeConstruct("Both", True, False)
        self.missed = FakeConstruct("Neither")
    
    def _makeOne(self, working_directory, metadata, options=None):
        if options is None:
            options = object()
        return ExecutionReport(working_directory, metadata, options)
        
    def test_header(self):
        expected_header = """
===============================================
Instrumental Condition/Decision Coverage Report
===============================================
"""
        
        reporter = self._makeOne('.', {})
        assert expected_header in reporter.report(), reporter.report()
    
    def test_showall_true(self):
        reporter = self._makeOne('.', {'somemodule': FakeMetadata({'1.1': self.missed})})
        assert "ConstructResult(Neither)" in reporter.report(True), reporter.report(True)




def test_func(emitt):    
    flag_1 = False
    if flag_1: # pragma: no cond(T)
        emit('flag_1 T')
    
    for flag_2 in [True, False]:
        if flag_2 and flag_1: # pragma: no cond(T T,F T)
            emit('flag_2 and flag_1 T T')



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import os
import sys

from instrumental.constructs import BooleanDecision

def _package(modulename):
    modulepath = '%s.py' % modulename.replace('.', os.path.sep)
    if any(os.path.exists(os.path.join(path, modulepath))
           for path in sys.path):
        return '.'.join(modulename.split('.')[:-1])
    else:
        return modulename

class BaseExecutionSummary(object):
    
    def __init__(self, conditions, statements, options):
        self.conditions = conditions
        self.statements = statements
        self.options = options
        
    @property
    def decisions(self):
        return [condition for condition in self.conditions
                if condition.is_decision()]
    
    def condition_rate(self):
        total_conditions = sum(condition.number_of_conditions(self.options.report_conditions_with_literals)
                               for condition in self.conditions)
        if not total_conditions:
            return 1.0
        hit_conditions = sum(condition.number_of_conditions_hit()
                             for condition in self.conditions)
        return hit_conditions / float(total_conditions)
    
    def decision_rate(self):
        total_conditions = sum(decision.number_of_conditions(self.options.report_conditions_with_literals)
                               for decision in self.decisions)
        if not total_conditions:
            return 1.0
        hit_conditions = sum(decision.number_of_conditions_hit()
                             for decision in self.decisions)
        return hit_conditions / float(total_conditions)
    
    def statement_rate(self):
        all_statements = []
        for modulename, statement_dict in self.statements.items():
            all_statements += statement_dict.items()
        
        total_statements = len(all_statements)
        if not total_statements:
            return 1.0
        hit_statements = sum(hit for (lineno, hit) in all_statements)
        return hit_statements / float(total_statements)


class ExecutionSummary(BaseExecutionSummary):
    
    def __init__(self, conditions, statements, options):
        self.statements = statements
        self.conditions = []
        for modulename in statements:
            self.conditions += conditions[modulename].values()
        self.options = options
        self._packages = None
    
    @property
    def packages(self):
        if self._packages is None:
            _statements = {}
            _conditions = {}
            for modulename in self.statements:
                _package_name = _package(modulename)
                
                _package_statements = \
                    _statements.setdefault(_package_name, {})
                _package_statements[modulename] = self.statements[modulename]
                
                _package_conditions = \
                    _conditions.setdefault(_package_name, [])
                for condition in self.conditions:
                    if condition.modulename == modulename:
                        _package_conditions.append(condition)
            
            self._packages = \
                dict((packagename, 
                      PackageExecutionSummary(packagename,
                                              _conditions[packagename],
                                              _statements[packagename],
                                              self.options))
                     for packagename in _statements)
        return self._packages
    
class PackageExecutionSummary(BaseExecutionSummary):
    
    def __init__(self, name, conditions, statements, options):
        super(PackageExecutionSummary, self).__init__(conditions, statements, options)
        self.name = name
        self._modules = None
    
    @property
    def modules(self):
        if self._modules is None:
            self._modules = {}
            for modulename in self.statements:
                _conditions = [condition for condition in self.conditions
                               if condition.modulename == modulename]
                self._modules[modulename] = (
                    ModuleExecutionSummary(modulename,
                                           _conditions,
                                           self.statements[modulename],
                                           self.options))
        return self._modules
    

class ModuleExecutionSummary(BaseExecutionSummary):
    
    def __init__(self, name, conditions, statements, options):
        super(ModuleExecutionSummary, self).__init__(conditions, statements, options)
        self.name = name
    
    def statement_rate(self):
        total_statements = len(self.statements)
        if not total_statements:
            return 1.0
        hit_statements = sum(hit for hit in self.statements.values())
        return hit_statements / float(total_statements)
    
